// TypedData gets converted to ArrayBuffer in Dartium.
function setTypedData(arrayBuffer, offset) {
  if (typeof(arrayBuffer.buffer) !== 'undefined') {
    arrayBuffer = arrayBuffer.buffer;
  }
  var bytes = new Uint8Array(arrayBuffer);
  Module['HEAPU8'].set(bytes, offset);
}
Module['setTypedData'] = setTypedData;

function getFloat64Array(ptr, length) {
  var bufView = new Float64Array(Module['buffer'], ptr, length);
  return bufView;
}
Module['getFloat64Array'] = getFloat64Array;

if (this.document === undefined) {
  // Web Worker mode
  
  // Incoming message = [ handle, [ command, args... ] ]
  // Outgoing message = [ handle, output ]
  // Outgoing message = [ handle, [ "error", error ] ]

  // handle is a random Number (assigned by the host) to match 
  // request and response.

  var s = 'number';
  var argTypes13 = [s, s, s, s, s, s, s, s, s, s, s, s, s];

  function copy_buf(/*TypedData*/ dat) {
    var ptr = Module['_malloc'](dat.length*dat.BYTES_PER_ELEMENT);
    var bytes = new Uint8Array(dat.buffer);
    Module['HEAPU8'].set(bytes, ptr);
    return ptr;
  }

  this.onmessage = function (event) {
    if (!Array.isArray(event["data"])) {
      postMessage("internal error: expected array");
      return;
    }

    var handle = event["data"][0];
    var payload = event["data"][1];
    if (handle == undefined && payload == undefined) {
      postMessage("internal error: no payload");
      return;
    }

    var iopt, n, nnz, nrhs, p_values, p_rowind, p_colptr, p_b, ldb,
      pp_factors, permc_spec, trans, p_info;
    var result;
    
    var i32 = Int32Array.BYTES_PER_ELEMENT; // pointers are 32-bit
    
    var p_info = Module['_malloc'](i32);
    Module['setValue'](p_info, 0, 'i32');

    var command = payload[0];
    switch (command) {
      case "dlinsol.factor":
      case "zlinsol.factor":
        var _values = payload[1];
        var _rowind = payload[2];
        var _colptr = payload[3];
        permc_spec = payload[4];

        var values = new Float64Array(_values);
        var rowind = new Int32Array(_rowind);
        var colptr = new Int32Array(_colptr);

        iopt = 1;
        n = colptr.length - 1;
        nnz = values.length;
        nrhs = 0;

        p_values = copy_buf(values);
        p_rowind = copy_buf(rowind);
        p_colptr = copy_buf(colptr);

        p_b = 0; // NULL
        ldb = 0;
        trans = 0;

        pp_factors = Module['_malloc'](i32);
        break;
      case "dlinsol.solve":
      case "zlinsol.solve":
        pp_factors = payload[1];
        n = payload[2];
        nrhs = payload[3];
        var _rhs = payload[4];
        ldb = payload[5];
        trans = payload[6];

        rhs = new Float64Array(_rhs);

        iopt = 2;
        nnz = 0;
        p_values = 0;
        p_rowind = 0;
        p_colptr = 0;

        p_b = copy_buf(rhs);
        if (ldb == undefined) {
          ldb = n;
        }

        permc_spec = 0;
        break;
      case "dlinsol.free":
      case "zlinsol.free":
        pp_factors = payload[1];

        iopt = 3;
        n = 0;
        nnz = 0;
        nrhs = 0;
        p_values = 0;
        p_rowind = 0;
        p_colptr = 0;
        p_b = 0;
        ldb = 0;
        permc_spec = 0;
        trans = 0;
        break;
      default:
        result = ["error", "invalid command: " + command];
    }

    switch (command) {
      case "dlinsol.factor":
      case "dlinsol.solve":
      case "dlinsol.free":
        Module.ccall('dlinsol', null, argTypes13,
          [iopt, n, nnz, nrhs, p_values, p_rowind, p_colptr, p_b, ldb,
           pp_factors, permc_spec, trans, p_info]);
        break;
      case "zlinsol.factor":
      case "zlinsol.solve":
      case "zlinsol.free":
        Module.ccall('zlinsol', null, argTypes13,
          [iopt, n, nnz, nrhs, p_values, p_rowind, p_colptr, p_b, ldb,
           pp_factors, permc_spec, trans, p_info]);
        break;
    }

    switch (command) {
      case "dlinsol.factor":
      case "zlinsol.factor":
        Module['_free'](p_values);
        Module['_free'](p_rowind);
        Module['_free'](p_colptr);
        result = pp_factors;
        break;
      case "dlinsol.solve":
      case "zlinsol.solve":
        var bview = new Float64Array(Module['buffer'], p_b, rhs.length);
        rhs.set(bview);
        result = rhs;
        Module['_free'](p_b);
        break;
      case "dlinsol.free":
      case "zlinsol.free":
        result = null;
    }

    var info = Module['getValue'](p_info, 'i32');
    Module['_free'](p_info);
    if (info != 0) {
      result = ["error", info];
    }

    var msg = [handle, result];
    this.postMessage(msg);
  }
}
